/* $Header: /home/hiro/projects/dvr2/PT1-arib25/arib25/src/RCS/bcc.c,v 1.2 2012/07/10 23:56:32 hiro Exp $
 *
 * bcc.c - BonCasClient
 *
 * Replaces major SCard API functions used in libarib25 with BonCasClient
 * counterparts
 *
 * Copyright (c) 2011 Hiro Sugawara
 */

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pcsclite.h>
#include <PCSC/winscard.h>

static const char *boncas = "BonCasClient at ";
static char cardreader[256];
static char *server;

static int checkBCC(void)
{
	if (server != NULL);
		return 1;
	server = getenv("BONCASSERVER");
	return (server != NULL);
}

static int openbcc(const char *server)
{
	int sock = -1;
	struct hostent *h;
	struct sockaddr_in sa;
	int port = 6900;
	char *p = index(server, ':');

	if (p) {
		*p++ = '\0';
		port = atoi(p);
		if (port <= 0)
			return -1;
	}

	h = gethostbyname(server);
	if (!h)
		return -1;
	if (p)
		*--p = ':';

	memset(&sa, 0, sizeof sa);
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	memcpy(&sa.sin_addr.s_addr, h->h_addr, h->h_length);

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0)
		return sock;
	if (connect(sock, &sa, sizeof sa) < 0) {
		close(sock);
		sock = -1;
	}
	return sock;
}

PCSC_API LONG SCardTransmit_bcc(SCARDHANDLE hCard,
	LPCSCARD_IO_REQUEST pioSendPci,
	LPCBYTE pbSendBuffer, DWORD cbSendLength,
	/*@out@*/ LPSCARD_IO_REQUEST pioRecvPci,
	/*@out@*/ LPBYTE pbRecvBuffer, LPDWORD pcbRecvLength)
{
#define BUF_MAX 254
	char sbuf[BUF_MAX + 2], rbuf[BUF_MAX + 2];
	int retry = 10;

	if (!checkBCC())
		return SCardTransmit(hCard, pioSendPci,
				pbSendBuffer, cbSendLength,
				pioRecvPci,
				pbRecvBuffer, pcbRecvLength);


	if (!cbSendLength) {
		*pcbRecvLength = 0;
		return SCARD_S_SUCCESS;
	}
	if (cbSendLength > BUF_MAX)
		return SCARD_E_NO_MEMORY;
	sbuf[0] = cbSendLength;
	memcpy(&sbuf[1], pbSendBuffer, cbSendLength);

	while (retry-- > 0) {
		if (send(hCard, sbuf, cbSendLength + 1, 0) != cbSendLength + 1)
			goto fail;
		if (recv(hCard, rbuf, sizeof sbuf, 0) > 0)
			break;
fail:
		sleep(1);
	}

	if (!retry)
		return SCARD_F_COMM_ERROR;

	*pcbRecvLength = (unsigned int)rbuf[0];
	memcpy(pbRecvBuffer, &rbuf[1], *pcbRecvLength);
	return SCARD_S_SUCCESS;
}

PCSC_API LONG SCardConnect_bcc(SCARDCONTEXT hContext,
	LPCSTR szReader,
	DWORD dwShareMode,
	DWORD dwPreferredProtocols,
	/*@out@*/ LPSCARDHANDLE phCard,
	/*@out@*/ LPDWORD pdwActiveProtocol)
{
	int sock;

	if (!checkBCC())
		return SCardConnect(hContext, szReader, dwShareMode,
					dwPreferredProtocols, phCard, pdwActiveProtocol);
	sock = openbcc(server);
	if (sock < 0)
		return SCARD_F_COMM_ERROR;
	*phCard = sock;
	return SCARD_S_SUCCESS;
}


PCSC_API LONG SCardDisconnect_bcc(SCARDHANDLE hCard, DWORD dwDisposition)
{
	if (!checkBCC())
		return SCardDisconnect(hCard, dwDisposition);

	close(hCard);
	return SCARD_S_SUCCESS;
}

PCSC_API LONG SCardEstablishContext_bcc(DWORD dwScope,
	/*@null@*/ LPCVOID pvReserved1, /*@null@*/ LPCVOID pvReserved2,
	/*@out@*/ LPSCARDCONTEXT phContext)
{
	if (!checkBCC())
		return SCardEstablishContext(dwScope, pvReserved1, pvReserved2, phContext);
	return SCARD_S_SUCCESS;
}

PCSC_API LONG SCardListReaders_bcc(SCARDCONTEXT hContext,
	/*@null@*/ /*@out@*/ LPCSTR mszGroups,
	/*@null@*/ /*@out@*/ LPSTR mszReaders,
	/*@out@*/ LPDWORD pcchReaders)
{
	if (!checkBCC())
		return SCardListReaders(hContext, mszGroups, mszReaders, pcchReaders);

	sprintf(cardreader, "%s %s", boncas, server);
	cardreader[strlen(cardreader) + 1] = '\0';
	*pcchReaders = strlen(cardreader) + 1;
	if (mszReaders)
		strcpy(mszReaders, cardreader);
	return SCARD_S_SUCCESS;
}

PCSC_API LONG SCardReleaseContext_bcc(SCARDCONTEXT hContext)
{
	if (!checkBCC())
		return SCardReleaseContext(hContext);
	return SCARD_S_SUCCESS;
}
